"""
====================
Minimalistic example
====================

Demonstrate basic use of :class:`coroica.CoroICA`
"""
import numpy as np
from coroica import CoroICA, UwedgeICA
from matplotlib import pyplot as plt
from sklearn.decomposition import FastICA
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import cross_val_predict
from sklearn.pipeline import Pipeline

# generate data
X = np.random.randn(500, 25)
y = np.random.randn(500,)

group_index = np.zeros(500,)
group_index[150:] = 1

X[:, :10] = X[:, :10] + 2 * y.reshape(-1, 1)

X[:150, 5:20] += 3 * np.random.randn(150, 15).dot(np.random.randn(15, 15))
X[150:, 5:20] += 5 * np.random.randn(350, 15).dot(np.random.randn(15, 15))

# define coroICA-based pipeline
model_coroICA = Pipeline(steps=[
    ('coroICA', CoroICA(n_components=10,
                        timelags=[5, 10],
                        max_matrices='no_partitions',
                        pairing='allpairs')),
    ('regression', LinearRegression())])

# get cross-validated predictions with coroICA-based pipeline
y_hat_coroICA = cross_val_predict(
    model_coroICA,
    X,
    y,
    fit_params={'coroICA__group_index': group_index})

# define uwedgeICA-based pipeline (second-order-based, ignores groupstructure)
model_uwedgeICA = Pipeline(steps=[
    ('uwedgeICA', UwedgeICA(n_components=10)),
    ('regression', LinearRegression())])

# get cross-validated predictions with uwedgeICA-based pipeline
y_hat_uwedgeICA = cross_val_predict(
    model_uwedgeICA,
    X,
    y)

# define pooled fastica-based pipeline (ignores groupstructure)
model_fastica = Pipeline(steps=[
    ('fastica', FastICA(n_components=10)),
    ('regression', LinearRegression())])

# get cross-validated predictions with pooled fastica-based pipeline
y_hat_fastica = cross_val_predict(
    model_fastica,
    X,
    y)

# for comparison plot scatter of predictions against the true y
plt.plot(y, y_hat_coroICA,
         '.',
         label='coroICA (correlation with true y {:.2f})'.format(
             np.corrcoef(y, y_hat_coroICA)[0, 1]))

plt.plot(y, y_hat_uwedgeICA,
         '.',
         label='uwedgeICA (correlation with true y {:.2f})'.format(
             np.corrcoef(y, y_hat_uwedgeICA)[0, 1]))

plt.plot(y, y_hat_fastica,
         '.',
         label='pooled fastica (correlation with true y {:.2f})'.format(
             np.corrcoef(y, y_hat_fastica)[0, 1]))

plt.title('y vs y_hat')
plt.legend(loc='best')
plt.grid(True)
plt.xlabel('true y')
plt.ylabel('y_hat')

plt.show()



from coroica import CoroICA, UwedgeICA
from sklearn.utils.estimator_checks import check_estimator


def test_transformer_coroICA():
    return check_estimator(CoroICA)


def test_transformer_uwedgeICA():
    return check_estimator(UwedgeICA)



# coding: utf-8
"""
coroICA
Implementation of the coroICA algorithm presented in
coroICA: Independent component analysis for grouped data
N Pfister*, S Weichwald*, P Bühlmann, B Schölkopf
https://arxiv.org/abs/1806.01094
"""
from .utils import autocov, rigidgroup
from .uwedge import uwedge
import itertools
import numpy as np
from scipy import linalg
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_X_y, check_array
from sklearn.utils.random import check_random_state
from sklearn.utils.validation import check_is_fitted
import warnings


class CoroICA(BaseEstimator, TransformerMixin):
    """coroICA transformer

    Parameters
    ----------
    n_components : int, optional
        Number of components to extract. If none is passed, the same number of
        components as the input has dimensions is used.
    n_components_uwedge : int, optional
        Number of components to extract during uwedge approximate joint
        diagonalization of the matrices. If none is passed, the same number of
        components as the input has dimensions is used.
    rank_components : boolean, optional
        When true, the components will be ordered in decreasing stability.
    pairing : {'complement', 'allpairs', 'neighbouring'}
        Whether difference matrices should be computed for all pairs of
        partition covariance matrices or only in a one-vs-complement scheme or
        only of neighbouring partition covariance matrices.
    max_matrices : float or 'no_partitions', optional (default=1)
        The fraction of (lagged) covariance matrices to use during training
        or, if 'no_partitions', at most as many covariance matrices are used
        as there are partitions.
    groupsize : int, optional
        Approximately how many samples, when doing a rigid grid, shall be in
        each group. If none is passed, all samples will be in one group unless
        group_index is passed during fitting in which case the provided group
        index is used (the latter is the advised and preferred way).
    partitionsize : int or list of int, optional
        Approximately how many samples, when doing a rigid grid, should be in
        each partition. If none is passed, a (hopefully sane) default is used
        unless partition_index is passed during fitting in which case
        the provided partition index is used.
    partitionsize : int, optional
        Approximately how many samples, when doing a rigid grid, should be in
        each partition. If none is passed, a (hopefully sane) default is used,
        again, unless partition_index is passed during fitting in which case
        the provided partition index is used.
    timelags : list of ints, optional
        List of time lags to be considered for computing lagged covariance
        matrices.
    instantcov : boolean, optional
        If False, no non-lagged instant (lag = 0) covariance matrices are used.
    max_iter : int, optional
        Maximum number of iterations for the uwedge approximate joint
        diagonalisation during fitting.
    tol : float, optional
        Tolerance for terminating the uwedge approximate joint diagonalisation
        during fitting.
    minimize_loss : boolean, optional
        If True at each iteration the loss of the uwedge approximate joint
        diagonalisation is computed (computationally expensive) and after
        convergence the V with minimal loss along the optimisation path is
        returned instead of the terminal V.
    condition_threshold : int, optional (default=None)
        If int, the uwedge iteration is stopped when the condition number of
        the unmixing matrix grows beyond condition_threshold. If None, no such
        condition number check is performed.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is seeded used by the random number generator;
        if RandomState instance, random_state is the random number generator;
        if None, the random number generator is the RandomState instance used
        by np.random.

    Attributes
    ----------
    V_ : array, shape (n, n_features)
        The unmixing matrix; where n=n_features if n_components and
        n_components_uwedge are None, n=n_components_uwedge if n_components is
        None, and n=n_components otherwise.
    converged_ : boolean
        Whether the approximate joint diagonalisation converged due to tol.
    n_iter_ : int
        Number of iterations of the approximate joint diagonalisation.
    meanoffdiag_ : float
        Mean absolute value of the off-diagonal values of the to be jointly
        diagonalised matrices, i.e., a proxy of the approximate joint
        diagonalisation objective function.
    """

    def __init__(self,
                 n_components=None,
                 n_components_uwedge=None,
                 rank_components=False,
                 pairing='complement',
                 max_matrices=1,
                 groupsize=None,
                 partitionsize=None,
                 timelags=None,
                 instantcov=True,
                 max_iter=5000,
                 tol=1e-12,
                 minimize_loss=False,
                 condition_threshold=None,
                 random_state=None):
        self.n_components = n_components
        self.n_components_uwedge = n_components_uwedge
        self.rank_components = rank_components
        self.pairing = pairing
        self.max_matrices = max_matrices
        self.groupsize = groupsize
        self.partitionsize = partitionsize
        self.timelags = timelags
        self.instantcov = instantcov
        self.max_iter = max_iter
        self.tol = tol
        self.minimize_loss = minimize_loss
        self.condition_threshold = condition_threshold
        self.random_state = random_state
        if self.timelags is None and not self.instantcov:
            warnings.warn('timelags=None and instantcov=True results in the '
                          'identity transformer, since no (lagged) covariance '
                          'matrices are to be diagonalised.',
                          UserWarning)

    def fit(self, X, y=None, group_index=None, partition_index=None):
        """Fit the model

        Parameters
        ----------
        X : array, shape (n_samples, n_features)
            where n_samples is the number of samples and
            n_features is the number of features.
        y : Ignored.
        group_index : array, optional, shape (n_samples,)
            Codes for each sample which group it belongs to; if no group index
            is provided a rigid grid with self.groupsize samples per
            group is used (which defaults to all samples if self.groupsize
            was not set).
        partition_index : array, optional, shape (n_samples,)
            Codes for each sample which partition it belongs to; if no
            partition index is provided a rigid grid with self.partitionsize
            samples per partition within each group is used (which has a
            (hopefully sane) default if self.partitionsize was not set).

        Returns
        -------
        self : object
            Returns self.
        """
        X = check_array(X, ensure_2d=True)

        n, dim = X.shape

        if n <= 1 or dim <= 1:
            self.V_ = np.eye(dim)
            return self

        random_state = check_random_state(self.random_state)

        # generate group and partition indices as needed
        if group_index is None and self.groupsize is None:
            group_index = np.zeros(n)
        elif group_index is None:
            group_index = rigidgroup(n, self.groupsize)

        # generate partition index as needed
        if partition_index is None and self.partitionsize is None:
            smallest_group = np.min([(group_index == group).sum()
                                     for group in np.unique(group_index)])
            partition_indices = [rigidpartition(
                group_index,
                np.max([dim, smallest_group // 2]))]
        elif partition_index is None and type(self.partitionsize) == list:
            partition_indices = [rigidpartition(group_index, partsize)
                                 for partsize in self.partitionsize]
        elif partition_index is None:
            partition_indices = [
                rigidpartition(group_index, self.partitionsize)]
        else:
            partition_indices = [partition_index]

        X, group_index = check_X_y(X, group_index)
        for partition_index in partition_indices:
            X, partition_index = check_X_y(X, partition_index)

        X = X.T

        no_groups = np.unique(group_index).shape[0]

        # computing covariance difference matrices
        timelags = []
        if self.instantcov:
            timelags.append(0)
        if self.timelags is not None:
            timelags.extend(self.timelags)
        no_timelags = len(timelags)
        for partition_index in partition_indices:
            if self.pairing == 'complement':
                if self.max_matrices == 'no_partitions':
                    max_matrices = 1.0
                else:
                    max_matrices = self.max_matrices
                no_pairs = 0
                for group in np.unique(group_index):
                    no_pairs += len(
                        np.unique(partition_index[group_index == group]))
                covmats = np.empty((no_pairs * no_timelags, dim, dim))
                idx = 0
                for group in np.unique(group_index):
                    unique_partitions = np.unique(
                        partition_index[group_index == group])
                    unique_partitions = random_state.choice(
                        unique_partitions,
                        size=np.ceil(
                            max_matrices * unique_partitions.shape[0]
                        ).astype(int),
                        replace=False)
                    for partition in unique_partitions:
                        ind1 = ((partition_index == partition) &
                                (group_index == group))
                        ind2 = ((partition_index != partition) &
                                (group_index == group))
                        if ind2.sum() > 0:
                            for timelag in timelags:
                                covmats[idx, :, :] = autocov(
                                    X[:, ind1], lag=timelag) - \
                                    autocov(X[:, ind2], lag=timelag)
                                idx += 1
                        else:
                            warnings.warn('Removing group {} since the '
                                          'partition is trivial, i.e., '
                                          'contains only exactly one '
                                          'set'.format(group),
                                          UserWarning)
            elif self.pairing in ['allpairs', 'neighbouring']:
                no_pairs = 0
                pairs_per_group = [None] * no_groups
                for i, group in enumerate(np.unique(group_index)):
                    unique_partitions = np.unique(
                        partition_index[group_index == group])
                    if self.pairing == 'allpairs':
                        pairs = np.asarray(list(
                            itertools.combinations(unique_partitions, 2)))
                    elif self.pairing == 'neighbouring':
                        pairs = np.asarray([
                            [unique_partitions[k], unique_partitions[k + 1]]
                            for k in range(len(unique_partitions) - 1)])
                    if pairs.shape[0] == 0:
                            warnings.warn('Removing group {} since the '
                                          'partition is trivial, i.e., '
                                          'contains only exactly one '
                                          'set'.format(group),
                                          UserWarning)
                    else:
                        if self.max_matrices == 'no_partitions':
                            max_matrices = (len(unique_partitions) - 1) / \
                                pairs.shape[0]
                        else:
                            max_matrices = self.max_matrices
                        pairs_per_group[i] = pairs[random_state.choice(
                            pairs.shape[0],
                            size=np.ceil(
                                max_matrices * pairs.shape[0]
                            ).astype(int),
                            replace=False
                        )]
                        no_pairs += pairs_per_group[i].shape[0]
                covmats = np.empty((no_pairs * no_timelags, dim, dim))
                idx = 0
                for pairs, group in zip(
                        pairs_per_group, np.unique(group_index)):
                    if pairs is not None:
                        for i, j in pairs:
                            ind1 = ((partition_index == i) &
                                    (group_index == group))
                            ind2 = ((partition_index == j) &
                                    (group_index == group))
                            for timelag in timelags:
                                covmats[idx, :, :] = autocov(
                                    X[:, ind1], lag=timelag) - \
                                    autocov(X[:, ind2], lag=timelag)
                                idx += 1
        covmats = covmats[:idx, :, :]

        Rx0 = np.cov(X)

        # joint diagonalisation
        self.V_, self.converged_, self.n_iter_, self.meanoffdiag_ = uwedge(
            covmats,
            Rx0=Rx0,
            eps=self.tol,
            minimize_loss=self.minimize_loss,
            n_iter_max=self.max_iter,
            n_components=self.n_components_uwedge,
            condition_threshold=self.condition_threshold)

        # normalise V
        normaliser = np.diag(self.V_.dot(Rx0.dot(self.V_.T)))
        self.V_ = self.V_ / (
            np.sign(normaliser) * np.sqrt(np.abs(normaliser)))[:, None]

        # rank components
        if self.rank_components or self.n_components is not None:
            A = linalg.pinv(self.V_)
            colcorrs = np.zeros(self.V_.shape[0])
            # running average
            for k in range(covmats.shape[0]):
                colcorrs += np.abs(np.corrcoef(
                    A.T,
                    self.V_.dot(covmats[k, ...].T)
                )[:dim, dim:].diagonal() / covmats.shape[0])
            sorting = np.argsort(colcorrs)[::-1]
            self.V_ = self.V_[sorting, :]

        if self.n_components is not None:
            self.V_ = self.V_[:self.n_components, :]

        return self

    def transform(self, X):
        """Returns the data projected onto the fitted components

        Parameters
        ----------
        X : array, shape (n_samples, n_features)
            where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transformed : array, shape (n_samples, n_components)
        """
        check_is_fitted(self, ['V_'])
        X = check_array(X)
        return self.V_.dot(X.T).T


def rigidpartition(group, nosamples):
    partition = np.zeros(group.shape)
    for e in np.unique(group):
        partition[np.where(group == e)] = rigidgroup(
            (group == e).sum(),
            nosamples) + partition.max() + 1
    return partition



import itertools
import numpy as np
import scipy.linalg as lin
import warnings


def uwedge(Rx,
           init=None,
           Rx0=None,
           return_diagonals=False,
           eps=1e-10,
           n_iter_max=1000,
           minimize_loss=False,
           verbose=False,
           n_components=None,
           condition_threshold=None):
    # Input:
    # Output:
    #   if return_diagonals=True
    #     V, diagonals, converged, iteration, meanoffdiag
    #   otherwise
    #     V, converged, iteration, meanoffdiag
    # Reference:
    # Fast Approximate Joint Diagonalization Incorporating Weight Matrices.
    # IEEE Transactions on Signal Processing, 2009.

    # 0) Preprocessing

    M, d = Rx.shape[0:2]
    if Rx0 is None:
        Rx0 = np.copy(Rx[0, :, :])

    if n_components is None:
        n_components = d

    # Initial guess
    if init is None and n_components == d:
        if Rx.shape[0] > 0:
            E, H = lin.eigh(Rx0)
            V = np.dot(np.diag(1. / np.sqrt(np.abs(E))), H.T)
        else:
            V = np.eye(d)
    elif init is None:
        E, H = lin.eigh(Rx0)
        mat = np.hstack([np.diag(1. / np.sqrt(np.abs(E[:n_components]))),
                         np.zeros((n_components, d - n_components))])
        V = np.dot(mat, H.T)
    else:
        V = init[:n_components, :]

    V = V / lin.norm(V, axis=1)[:, None]

    current_best = [None, np.inf, 0, None]

    for iteration in itertools.count():
        # 1) Generate Rs
        Rs = np.stack([V.dot(Rxx.dot(V.T)) for Rxx in Rx])

        # 2) Use Rs to construct A, equation (24) in paper with W=Id
        # 3) Set A1=Id and substitute off-diagonals
        Rsdiag = Rs.diagonal(axis1=1, axis2=2)
        Rsdiagprod = Rsdiag.T.dot(Rsdiag)
        denom_mat = np.outer(
            Rsdiagprod.diagonal(),
            Rsdiagprod.diagonal()) - Rsdiagprod**2
        Rkl = np.einsum('ill,ikl->kl', Rs, Rs)
        num_mat = Rsdiagprod.diagonal()[:, None] * Rkl - Rsdiagprod * Rkl.T
        denom_mat[denom_mat == 0] = np.finfo(V.dtype).eps
        A = num_mat / (denom_mat + np.eye(n_components))
        np.fill_diagonal(A, 1)

        # 4) Set new V
        Vold = np.copy(V)
        V = lin.lstsq(A, Vold,
                      check_finite=False,
                      lapack_driver='gelsy')[0]

        # 5) Normalise V
        V = V / lin.norm(V, axis=1)[:, None]

        if minimize_loss:
            diagonals = np.stack([V.dot(Rxx.dot(V.T)) for Rxx in Rx])
            meanoffdiag = np.mean(
                diagonals[:, ~np.eye(n_components, dtype=bool)]**2)
            if meanoffdiag < current_best[1]:
                current_best = [V, meanoffdiag, iteration, diagonals]

        # 6) Check convergence
        changeinV = np.max(np.abs(V - Vold))
        if iteration >= n_iter_max - 1:
            converged = False
            break
        elif changeinV < eps:
            converged = True
            break

        # 7) Check condition number
        if (condition_threshold is not None
                and np.linalg.cond(V) > condition_threshold):
            converged = False
            V = Vold
            iteration -= 1
            warnings.warn('Abort uwedge due to unreasonably growing condition '
                          'number of unmixing matrix V',
                          UserWarning)
            break

    # Rescale
    if minimize_loss:
        V, meanoffdiag, iteration, diagonals = current_best
    else:
        diagonals = np.stack([V.dot(Rxx.dot(V.T)) for Rxx in Rx])
        meanoffdiag = np.mean(
            diagonals[:, ~np.eye(n_components, dtype=bool)]**2)

    # Return
    if return_diagonals:
        return V, diagonals, converged, iteration, meanoffdiag
    return V, converged, iteration, meanoffdiag



import numpy as np


def autocov(X, lag=0):
    if lag == 0:
        return np.cov(X)
    else:
        A = X[:, lag:] - X[:, lag:].mean(axis=1, keepdims=True)
        B = X[:, :-lag] - X[:, :-lag].mean(axis=1, keepdims=True)
        n = A.shape[1] - 1
        return A.dot(B.T) / n


def rigidgroup(length, nosamples):
    groups = int(np.floor(length / nosamples))
    changepoints = [int(np.round(a))
                    for a in np.linspace(0, length, groups + 1)]
    changepoints = list(set(changepoints))
    changepoints.sort()
    index = np.zeros(length)
    for (i, a, b) in zip(range(groups), changepoints[:-1], changepoints[1:]):
        index[a:b] = i
    return index



# coding: utf-8
"""
uwedgeICA
Implementation of the NSS uwedgeICA algorithms used for comparison in
coroICA: Independent component analysis for grouped data
N Pfister*, S Weichwald*, P Bühlmann, B Schölkopf
https://arxiv.org/abs/1806.01094
"""
from .utils import autocov
from .utils import rigidgroup as rigidpartition
from .uwedge import uwedge
import numpy as np
from scipy import linalg
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_X_y, check_array
from sklearn.utils.validation import check_is_fitted
import warnings


class UwedgeICA(BaseEstimator, TransformerMixin):
    """uwedgeICA transformer

    Parameters
    ----------
    n_components : int, optional
        Number of components to extract. If none is passed, the same number of
        components as the input has dimensions is used.
    n_components_uwedge : int, optional
        Number of components to extract during uwedge approximate joint
        diagonalization of the matrices. If none is passed, the same number of
        components as the input has dimensions is used.
    rank_components : boolean, optional
        When true, the components will be ordered in decreasing stability.
    partitionsize : int or list of int, optional
        Approximately how many samples, when doing a rigid grid, should be in
        each partition. If none is passed, a (hopefully sane) default is used
        unless partition_index is passed during fitting in which case
        the provided partition index is used.
    timelags : list of ints, optional
        List of time lags to be considered for computing lagged covariance
        matrices.
    instantcov : boolean, optional
        If False, no non-lagged instant (lag = 0) covariance matrices are used.
    max_iter : int, optional
        Maximum number of iterations for the uwedge approximate joint
        diagonalisation during fitting.
    tol : float, optional
        Tolerance for terminating the uwedge approximate joint diagonalisation
        during fitting.
    minimize_loss : boolean, optional
        If True at each iteration the loss of the uwedge approximate joint
        diagonalisation is computed (computationally expensive) and after
        convergence the V with minimal loss along the optimisation path is
        returned instead of the terminal V.
    condition_threshold : int, optional (default=None)
        If int, the uwedge iteration is stopped when the condition number of
        the unmixing matrix grows beyond condition_threshold. If None, no such
        condition number check is performed.

    Attributes
    ----------
    V_ : array, shape (n, n_features)
        The unmixing matrix; where n=n_features if n_components and
        n_components_uwedge are None, n=n_components_uwedge if n_components is
        None, and n=n_components otherwise.
    converged_ : boolean
        Whether the approximate joint diagonalisation converged due to tol.
    n_iter_ : int
        Number of iterations of the approximate joint diagonalisation.
    meanoffdiag_ : float
        Mean absolute value of the off-diagonal values of the to be jointly
        diagonalised matrices, i.e., a proxy of the approximate joint
        diagonalisation objective function.
    """

    def __init__(self,
                 n_components=None,
                 n_components_uwedge=None,
                 rank_components=False,
                 partitionsize=None,
                 timelags=None,
                 instantcov=True,
                 max_iter=1000,
                 tol=1e-12,
                 minimize_loss=False,
                 condition_threshold=None):
        self.n_components = n_components
        self.n_components_uwedge = n_components_uwedge
        self.rank_components = rank_components
        self.partitionsize = partitionsize
        self.timelags = timelags
        self.instantcov = instantcov
        self.max_iter = max_iter
        self.tol = tol
        self.minimize_loss = minimize_loss
        self.condition_threshold = condition_threshold
        if self.timelags is None and not self.instantcov:
            warnings.warn('timelags=None and instantcov=True results in the '
                          'identity transformer, since no (lagged) covariance '
                          'matrices are to be diagonalised.',
                          UserWarning)

    def fit(self, X, y=None, partition_index=None):
        """Fit the model

        Parameters
        ----------
        X : array, shape (n_samples, n_features)
            where n_samples is the number of samples and
            n_features is the number of features.
        y : Ignored.
        partition_index : array, optional, shape (n_samples,)
            Codes for each sample which partition it belongs to; if no
            partition index is provided a rigid grid with self.partitionsize_
            samples per partition within each group is used (which has a
            (hopefully sane) default if self.partitionsize_ was not set).

        Returns
        -------
        self : object
            Returns self.
        """
        X = check_array(X, ensure_2d=True)

        n, dim = X.shape

        if (n <= 1
                or dim <= 1
                or (self.timelags is None and not self.instantcov)):
            self.V_ = np.eye(dim)
            return self

        # generate partition index as needed
        if partition_index is None and self.partitionsize is None:
            partition_indices = [rigidpartition(
                n,
                np.max([dim, n // 2]))]
        elif partition_index is None and type(self.partitionsize) == list:
            partition_indices = [rigidpartition(n, partsize)
                                 for partsize in self.partitionsize]
        elif partition_index is None:
            partition_indices = [rigidpartition(n, self.partitionsize)]
        else:
            partition_indices = [partition_index]

        for partition_index in partition_indices:
            X, partition_index = check_X_y(X, partition_index)

        X = X.T

        # computing covariance matrices
        no_partitions = 0
        for partition_index in partition_indices:
            no_partitions += len(np.unique(partition_index))
        timelags = []
        if self.instantcov:
            timelags.append(0)
        if self.timelags is not None:
            timelags.extend(self.timelags)
        no_timelags = len(timelags)
        covmats = np.empty((no_partitions * no_timelags, dim, dim))
        idx = 0
        for partition_index in partition_indices:
            for partition in np.unique(partition_index):
                ind = (partition_index == partition)
                for timelag in timelags:
                    covmats[idx, :, :] = autocov(X[:, ind], lag=timelag)
                    idx += 1

        Rx0 = np.cov(X)

        # joint diagonalisation
        self.V_, self.converged_, self.n_iter_, self.meanoffdiag_ = uwedge(
            covmats,
            Rx0=Rx0,
            eps=self.tol,
            minimize_loss=self.minimize_loss,
            n_iter_max=self.max_iter,
            n_components=self.n_components_uwedge,
            condition_threshold=self.condition_threshold)

        # normalise V
        normaliser = np.diag(self.V_.dot(Rx0.dot(self.V_.T)))
        self.V_ = self.V_ / (
            np.sign(normaliser) * np.sqrt(np.abs(normaliser)))[:, None]

        # rank components
        if self.rank_components or self.n_components is not None:
            A = linalg.pinv(self.V_)
            colcorrs = np.zeros(self.V_.shape[0])
            # running average
            for k in range(covmats.shape[0]):
                colcorrs += np.abs(np.corrcoef(
                    A.T,
                    self.V_.dot(covmats[k, ...].T)
                )[:dim, dim:].diagonal() / covmats.shape[0])
            sorting = np.argsort(colcorrs)[::-1]
            self.V_ = self.V_[sorting, :]

        if self.n_components is not None:
            self.V_ = self.V_[:self.n_components, :]

        return self

    def transform(self, X):
        """Returns the data projected onto the fitted components

        Parameters
        ----------
        X : array, shape (n_samples, n_features)
            where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transformed : array, shape (n_samples, n_components)
        """
        check_is_fitted(self, ['V_'])
        X = check_array(X)
        return self.V_.dot(X.T).T
